import { DocumentChangeType, ICollectionRef, IBulk } from "./types";
import * as Utils from "@hibas123/utils";
export interface Change {
    data: any;
    document: string;
    type: DocumentChangeType;
}
export declare type IWriteQueries = "set" | "update" | "delete" | "add";
export declare type ICollectionQueries = "get" | "add" | "keys" | "delete-collection" | "list";
export declare type IDocumentQueries = "get" | "set" | "update" | "delete";
export declare type ICallQueryTypes = IWriteQueries | ICollectionQueries | IDocumentQueries | "snapshot" | "unsubscribe";
interface ITypedQuery<T> {
    path: string[];
    type: T;
    data?: any;
    options?: any;
}
export declare type IQueryRequest = ITypedQuery<ICallQueryTypes> | ITypedQuery<"set" | "update" | "delete">[];
export declare abstract class Snapshot<T> {
    protected id: string;
    isSubscribed: boolean;
    abstract subscribe(): Promise<() => void>;
    abstract unsubscribe(): void;
    abstract receivedData(data: T): void;
    abstract resubscribe(): void;
    abstract destroy(): void;
}
export declare class OfflineError extends Error {
    constructor();
}
export interface IConnector {
    snapshots: Map<string, Snapshot<any>>;
    offline: Utils.AwaitStore<boolean>;
    queryRequest<T = any>(query?: IQueryRequest, id?: string, ns?: string): Promise<T>;
    close(): Promise<void>;
}
export default class Database {
    connector: IConnector;
    constructor(connector: IConnector);
    constructor(url: string, database: string, accesskey?: string, authkey?: string, rootkey?: string);
    collection<T = any>(name: string): ICollectionRef<T>;
    collections(): Promise<ICollectionRef[]>;
    bulk(): IBulk;
    close(): Promise<void>;
}
export {};
