"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _WebSocketConnector_offlineStore, _WebSocketConnector_closedStore, _WebSocketConnector_requests, _WebSocketConnector_request, _WebSocketConnector_url;
Object.defineProperty(exports, "__esModule", { value: true });
exports.OfflineError = exports.Snapshot = void 0;
const types_1 = require("./types");
const NanoID = require("nanoid");
const Utils = require("@hibas123/utils");
const WebSocket = require("isomorphic-ws");
const isEqual = require("lodash.isequal");
const delay = (time) => new Promise((yes) => setTimeout(yes, time));
const ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
const custNano = NanoID.customAlphabet(ALPHABET, 21);
const docNanoID = NanoID.customAlphabet(ALPHABET, 32);
class Snapshot {
    constructor() {
        this.id = custNano();
        this.isSubscribed = false;
    }
}
exports.Snapshot = Snapshot;
// const QueryRequest = Symbol("queryRequest");
// const Snapshots = Symbol("snapshots");
class OfflineError extends Error {
    constructor() {
        super("Offline");
    }
}
exports.OfflineError = OfflineError;
class WebSocketConnector {
    constructor(url, database, accesskey, authkey, rootkey) {
        _WebSocketConnector_offlineStore.set(this, new Utils.AwaitStore(true));
        _WebSocketConnector_closedStore.set(this, new Utils.AwaitStore(false));
        _WebSocketConnector_requests.set(this, new Map());
        _WebSocketConnector_request.set(this, new Utils.AsyncIteratorFromCB());
        this.snapshots = new Map();
        _WebSocketConnector_url.set(this, void 0);
        const u = new URL(url);
        if (authkey)
            u.searchParams.set("authkey", authkey);
        if (rootkey)
            u.searchParams.set("rootkey", rootkey);
        if (accesskey)
            u.searchParams.set("accesskey", accesskey);
        if (u.protocol === "http:") {
            u.protocol = "ws:";
        }
        else if (u.protocol === "https:") {
            u.protocol = "wss:";
        }
        u.searchParams.set("database", database);
        __classPrivateFieldSet(this, _WebSocketConnector_url, u, "f");
        this._connection();
    }
    get offline() {
        return {
            state: __classPrivateFieldGet(this, _WebSocketConnector_offlineStore, "f").value,
            ...__classPrivateFieldGet(this, _WebSocketConnector_offlineStore, "f").getPublicApi(),
        };
    }
    async _connection() {
        const socketStore = new Utils.AwaitStore(undefined);
        __classPrivateFieldGet(this, _WebSocketConnector_closedStore, "f").subscribe((val) => {
            var _a;
            if (val === true) {
                (_a = socketStore.value) === null || _a === void 0 ? void 0 : _a.close(); //TODO: Make gracefull shutdown
            }
        });
        (async () => {
            for await (const message of __classPrivateFieldGet(this, _WebSocketConnector_request, "f")) {
                if (__classPrivateFieldGet(this, _WebSocketConnector_closedStore, "f").value)
                    return;
                await socketStore.awaitValue((value) => value !== undefined);
                const data = new TextEncoder().encode(message);
                await new Promise((yes, no) => socketStore.value.send(data, (err) => (err ? no(err) : yes())));
            }
        })();
        while (true) {
            try {
                if (__classPrivateFieldGet(this, _WebSocketConnector_closedStore, "f").value)
                    return;
                const socket = new WebSocket(__classPrivateFieldGet(this, _WebSocketConnector_url, "f").href);
                await new Promise((yes, no) => {
                    socket.once("open", yes);
                    socket.once("error", no);
                });
                __classPrivateFieldGet(this, _WebSocketConnector_offlineStore, "f").send(false); // Online now
                socketStore.send(socket);
                socket.on("message", (data) => {
                    if (data instanceof ArrayBuffer) {
                        data = new TextDecoder().decode(data);
                    }
                    else if (data instanceof Buffer) {
                        data = data.toString("utf-8");
                    }
                    else if (Array.isArray(data)) {
                        data = Buffer.concat(data).toString("utf-8");
                    }
                    else {
                        let message = JSON.parse(data);
                        switch (message.ns) {
                            case "message":
                                this.handleResponse(message.data);
                                break;
                            case "snapshot":
                                this.handleSnapshot(message.data);
                                break;
                            case "error_msg":
                                // TODO: Some kind of error handling etc.
                                console.log("Server error:", message.data);
                                break;
                        }
                    }
                });
                socket.on("ping", (data) => {
                    socket.pong(data);
                });
                socket.on("pong", () => {
                    //TODO: Maybe add somthing here?
                });
                await new Promise((yes, no) => {
                    socket.once("close", () => {
                        if (__classPrivateFieldGet(this, _WebSocketConnector_closedStore, "f").value)
                            return;
                        yes();
                    });
                    socket.on("error", no);
                });
            }
            catch (err) {
                console.log(err);
            }
            await delay(1000);
        }
    }
    handleSnapshot({ id, data }) {
        let handler = this.snapshots.get(id);
        if (handler)
            handler.receivedData(data);
    }
    handleResponse({ id, error, data, }) {
        let req = __classPrivateFieldGet(this, _WebSocketConnector_requests, "f").get(id);
        if (req) {
            if (error)
                req(new Error(data), undefined);
            else
                req(undefined, data);
        }
        else {
            console.error("Invalid response", id);
        }
    }
    async queryRequest(query, id, ns = "v2") {
        if (!id)
            id = custNano();
        return new Promise(async (yes, no) => {
            let offlinePrms;
            const finish = (err, data) => {
                clearTimeout(to);
                err ? no(err) : yes(data);
                if (offlinePrms)
                    offlinePrms.ignore();
                // @ts-ignore
                __classPrivateFieldGet(this, _WebSocketConnector_requests, "f").delete(id);
            };
            let to = setTimeout(() => {
                finish(new Error("Timeout!"), undefined);
            }, 15000);
            await __classPrivateFieldGet(this, _WebSocketConnector_offlineStore, "f").awaitValue(false);
            // @ts-ignore
            offlinePrms = __classPrivateFieldGet(this, _WebSocketConnector_offlineStore, "f").awaitValue(true);
            // @ts-ignore
            offlinePrms.then(() => {
                offlinePrms = undefined;
                finish(new OfflineError(), undefined);
            });
            // @ts-ignore
            __classPrivateFieldGet(this, _WebSocketConnector_requests, "f").set(id, finish);
            __classPrivateFieldGet(this, _WebSocketConnector_request, "f").send(JSON.stringify({
                ns,
                data: { id, query },
            }));
        });
    }
    async close() {
        __classPrivateFieldGet(this, _WebSocketConnector_closedStore, "f").send(true);
        __classPrivateFieldGet(this, _WebSocketConnector_offlineStore, "f").close();
        __classPrivateFieldGet(this, _WebSocketConnector_request, "f").close();
    }
}
_WebSocketConnector_offlineStore = new WeakMap(), _WebSocketConnector_closedStore = new WeakMap(), _WebSocketConnector_requests = new WeakMap(), _WebSocketConnector_request = new WeakMap(), _WebSocketConnector_url = new WeakMap();
// TODO: Request Timeout
class Database {
    constructor(url, database, accesskey, authkey, rootkey) {
        if (typeof url === "string") {
            this.connector = new WebSocketConnector(url, database, accesskey, authkey, rootkey);
        }
        else {
            this.connector = url;
        }
    }
    collection(name) {
        return new CollectionRef(this, [name]);
    }
    async collections() {
        return this.connector
            .queryRequest({
            type: "list",
            path: [],
        })
            .then((res) => res.map((elm) => new CollectionRef(this, elm)));
    }
    bulk() {
        return new Bulk(this);
    }
    async close() {
        await this.connector.close();
    }
}
exports.default = Database;
class Query {
    constructor(db, prefix) {
        this.db = db;
        this.prefix = prefix;
        this._where = [];
    }
    _getNew() {
        let q = new Query(this.db, this.prefix);
        q._where = [...this._where];
        q._limit = this._limit;
        q._orderBy = this._orderBy;
        return q;
    }
    _getDocRef(id) {
        return new DocumentRef(this.db, [...this.prefix, id || docNanoID()]);
    }
    get database() {
        return this.db;
    }
    get path() {
        return [...this.prefix];
    }
    where(fieldPath, opStr, value) {
        let q = this._getNew();
        q._where.push({ fieldPath, opStr, value });
        return q;
    }
    orderBy(fieldPath, direction) {
        let q = this._getNew();
        q._orderBy = { fieldPath, direction };
        return q;
    }
    /**
     * WIP: the functionality is not stable yet!
     * @param limit The number of items
     */
    limit(limit) {
        let q = this._getNew();
        q._limit = limit;
        return q;
    }
    _getField(data, fieldpath) {
        let fields = fieldpath.split(".");
        let value = data;
        while (fields.length > 0) {
            if (!value)
                break;
            // @ts-ignore
            value = value[fields.shift()];
        }
        return value;
    }
    _order(data) {
        if (!this._orderBy)
            return data;
        data = data.sort((a, b) => {
            //@ts-ignore
            let val1 = this._getField(a.data(), this._orderBy.fieldPath);
            //@ts-ignore
            let val2 = this._getField(b.data(), this._orderBy.fieldPath);
            if (["number", "bigint", "boolean", "string"].indexOf(typeof val1) < 0)
                return 1;
            if (["number", "bigint", "boolean", "string"].indexOf(typeof val2) < 0)
                return 1;
            //TODO: Check how to resolve error without transpiler hacks
            //@ts-ignore
            return (val1 > val2) - (val2 > val1);
        });
        if (this._orderBy.direction === "desc")
            data = data.reverse();
        return data;
    }
    async get() {
        let data = this._order((await this.db.connector.queryRequest({
            type: "get",
            path: this.prefix,
            options: {
                limit: this._limit,
                where: this._where,
            },
        })).map((e) => {
            return {
                id: e.id,
                data: () => e.data,
                ref: this._getDocRef(e.id),
            };
        }));
        return {
            docs: data,
            docChanges: () => {
                return data.map((doc) => {
                    return {
                        doc,
                        type: "added",
                    };
                });
            },
            size: data.length,
            empty: data.length <= 0,
        };
    }
    async onSnapshot() {
        const asyncIter = new Utils.AsyncIteratorFromCB();
        await new Query.QuerySnapshot(this, asyncIter.getCallback()).subscribe();
        return asyncIter;
    }
}
Query.QuerySnapshot = class extends Snapshot {
    constructor(query, next) {
        super();
        this.query = query;
        this.next = next;
        this.stored = new Map();
    }
    async subscribe() {
        this.query.database.connector.snapshots.set(this.id, this);
        await this.resubscribe();
        return () => this.unsubscribe();
    }
    unsubscribe() {
        this.query.database.connector.snapshots.delete(this.id);
        return this.query.database.connector.queryRequest({
            type: "unsubscribe",
            path: this.query.path,
        }, this.id, "unsubscribe");
    }
    _change(changes) {
        let docs = Array.from(this.stored.values());
        docs = this.query._order(docs); //TODO: a little inefficient...
        this.next(undefined, {
            docs,
            docChanges: () => changes,
            empty: docs.length <= 0,
            size: docs.length,
        });
    }
    receivedData(data) {
        let changes = data.map((raw) => {
            let doc = {
                id: raw.id,
                data: () => raw.data,
                ref: this.query._getDocRef(raw.id),
            };
            if (raw.type === "deleted")
                this.stored.delete(raw.id);
            else
                this.stored.set(raw.id, doc);
            return {
                doc,
                type: raw.type,
            };
        });
        this._change(changes);
    }
    resubscribe() {
        if (this.isSubscribed)
            return;
        this.isSubscribed = true;
        return this.query.database.connector
            .queryRequest({
            type: "snapshot",
            options: {
                where: this.query._where,
                limit: this.query._limit,
            },
            path: this.query.path,
        }, this.id, "snapshot")
            .catch((err) => {
            this.query.database.connector.snapshots.delete(this.id);
            return Promise.reject(err);
        })
            .then((data) => {
            let newStored = new Map();
            let changes = data
                .map((e) => {
                let change = {
                    doc: {
                        id: e.id,
                        data: () => e.data,
                        ref: this.query._getDocRef(e.id),
                    },
                    type: "added",
                };
                newStored.set(change.doc.id, change.doc);
                let old = this.stored.get(e.id);
                if (old) {
                    this.stored.delete(change.doc.id);
                    if (isEqual(old, e.data)) {
                        return undefined;
                    }
                    else {
                        change.type = "modified";
                    }
                }
                return change;
            })
                .filter((e) => !!e);
            changes = [
                ...changes,
                ...Array.from(this.stored.values()).map((stored) => {
                    stored.data = () => null;
                    return {
                        doc: stored,
                        type: "deleted",
                    };
                }),
            ];
            this.stored = newStored;
            this._change(changes);
        })
            .catch((err) => {
            if (!(err instanceof OfflineError)) {
                this.next(err, undefined);
            }
            this.isSubscribed = false;
        });
    }
    destroy() {
        if (this.query.database.connector.snapshots.has(this.id)) {
            this.unsubscribe();
        }
        // @ts-ignore
        this.next = undefined;
        // @ts-ignore
        this.query = undefined;
        // @ts-ignore
        this.stored = undefined;
    }
};
class CollectionRef extends Query {
    get name() {
        return this.prefix[this.prefix.length - 1];
    }
    doc(id) {
        return this._getDocRef(id);
    }
    keys() {
        return this.db.connector.queryRequest({
            type: "keys",
            path: this.prefix,
        });
    }
    async add(data) {
        let id = await this.db.connector.queryRequest({
            type: "add",
            path: this.prefix,
            data,
        });
        return new DocumentRef(this.db, [...this.prefix, id]);
    }
    delete() {
        return this.db.connector.queryRequest({
            type: "delete-collection",
            path: this.prefix,
        });
    }
}
class DocumentRef {
    constructor(db, prefix) {
        this.db = db;
        this.prefix = prefix;
    }
    get id() {
        return this.prefix[this.prefix.length - 1];
    }
    get database() {
        return this.db;
    }
    get path() {
        return [...this.prefix];
    }
    get() {
        return this.db.connector.queryRequest({
            type: "get",
            path: this.prefix,
        });
    }
    _set(data) {
        return {
            type: "set",
            path: this.prefix,
            data,
        };
    }
    set(data) {
        return this.db.connector.queryRequest(this._set(data));
    }
    _update(update) {
        let u = {};
        for (const field in update) {
            const data = update[field];
            let res;
            if (typeof data === "object") {
                if (data[types_1.TimestampSymbol]) {
                    res = {
                        type: "timestamp",
                    };
                }
                else if (data[types_1.PushSymbol]) {
                    res = {
                        type: "push",
                        value: data.value,
                    };
                }
                else if (data[types_1.IncrementSymbol]) {
                    res = {
                        type: "increment",
                        value: data.value,
                    };
                }
                else {
                    res = {
                        type: "value",
                        value: data,
                    };
                }
            }
            else {
                res = {
                    type: "value",
                    value: data,
                };
            }
            u[field] = res;
        }
        return {
            type: "update",
            path: this.prefix,
            data: u,
        };
    }
    update(update) {
        return this.db.connector.queryRequest(this._update(update));
    }
    _delete() {
        return {
            type: "delete",
            path: this.prefix,
        };
    }
    delete() {
        return this.db.connector.queryRequest(this._delete());
    }
    async onSnapshot() {
        const asyncIter = new Utils.AsyncIteratorFromCB();
        await new DocumentRef.DocumentSnapshot(this, asyncIter.getCallback()).subscribe();
        return asyncIter;
    }
    collection(name) {
        return new CollectionRef(this.db, [...this.prefix, name]);
    }
}
DocumentRef.DocumentSnapshot = class extends Snapshot {
    constructor(document, next) {
        super();
        this.document = document;
        this.next = next;
        this.stored = undefined;
    }
    async subscribe() {
        this.document.database.connector.snapshots.set(this.id, this);
        await this.resubscribe();
        return () => this.unsubscribe();
    }
    unsubscribe() {
        this.document.database.connector.snapshots.delete(this.id);
        return this.document.database.connector.queryRequest({
            type: "unsubscribe",
            path: this.document.path,
        }, this.id, "unsubscribe");
    }
    _change() {
        this.next(undefined, {
            id: this.document.id,
            data: () => this.stored,
            ref: this.document,
        });
    }
    receivedData(changes) {
        changes.forEach((change) => {
            this.stored = change.data;
            this._change();
        });
    }
    resubscribe() {
        if (this.isSubscribed)
            return;
        this.isSubscribed = true;
        return this.document.database.connector
            .queryRequest({
            type: "snapshot",
            path: this.document.path,
        }, this.id, "snapshot")
            .catch((err) => {
            this.document.database.connector.snapshots.delete(this.id);
            return Promise.reject(err);
        })
            .then((data) => {
            this.isSubscribed = true;
            if (!isEqual(this.stored, data)) {
                this.stored = data;
                this._change();
            }
        })
            .catch((err) => {
            if (!(err instanceof OfflineError)) {
                this.next(err, undefined);
            }
            this.isSubscribed = false;
        });
    }
    destroy() {
        if (this.document.database.connector.snapshots.has(this.id)) {
            this.unsubscribe();
        }
        //@ts-ignore
        this.next = undefined;
        //@ts-ignore
        this.document = undefined;
        //@ts-ignore
        this.stored = undefined;
    }
};
class Bulk {
    constructor(_database) {
        this._database = _database;
        this._queries = [];
    }
    set(ref, data) {
        this._queries.push(ref._set(data));
        return this;
    }
    update(ref, updates) {
        this._queries.push(ref._update(updates));
        return this;
    }
    delete(ref) {
        this._queries.push(ref._delete());
        return this;
    }
    async commit() {
        if (this._queries.length > 0) {
            await this._database.connector.queryRequest(this._queries);
        }
    }
}
//# sourceMappingURL=database.js.map